import pytest

from yenta.tasks import (
    task, build_parameter_spec, TaskDef, InvalidTaskDefinitionError, ParameterSpec,
    ParameterType, ResultSpec, ResultType
)


def test_build_param_spec():

    @task
    def foo(previous_results):
        pass

    spec = build_parameter_spec(foo)
    expected_spec = [ParameterSpec('previous_results', ParameterType.PIPELINE_RESULTS)]

    assert(spec == expected_spec)

    @task
    def bar(x: 'foo__values__x', y: 'foo__artifacts__y'):
        pass

    spec = build_parameter_spec(bar)
    expected_spec = [
        ParameterSpec('x', ParameterType.EXPLICIT, ResultSpec('foo', ResultType.VALUE, 'x')),
        ParameterSpec('y', ParameterType.EXPLICIT, ResultSpec('foo', ResultType.ARTIFACT, 'y'))
    ]

    assert(spec == expected_spec)

    selector = lambda res: res
    selectors = {'x': selector}

    @task(selectors={'x': selector})
    def baz(x):
        pass

    spec = build_parameter_spec(baz, selectors)
    expected_spec = [ParameterSpec('x', ParameterType.EXPLICIT, None, selector)]

    assert(spec == expected_spec)


def test_invalid_param_spec():

    with pytest.raises(InvalidTaskDefinitionError) as ex:

        @task
        def foo(x: 'bar__values__x', y: int):
            pass

        _ = build_parameter_spec(foo)

    assert('Annotation string or selector missing' in str(ex.value))

    with pytest.raises(InvalidTaskDefinitionError) as ex:

        @task
        def foo(x: 'bar__values__x', y: 'bad_annotation'):
            pass

        _ = build_parameter_spec(foo)

    assert ('Invalid function annotation' in str(ex.value))


def test_task_definition():

    @task
    def foo(previous_results):
        pass

    expected_def = TaskDef('foo', None, True, [ParameterSpec('previous_results', ParameterType.PIPELINE_RESULTS)])

    assert(foo.task_def == expected_def)




from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Optional, Union

from yenta.utils.files import file_hash


@dataclass
class Artifact:

    location: Union[str, Path]
    date_created: str = None
    hash: Optional[str] = None
    meta: dict = None

    def artifact_hash(self):
        raise NotImplementedError

    def __post_init__(self):

        if not self.date_created:
            self.date_created = str(datetime.now())

    def __eq__(self, other):

        return self.location == other.location and self.hash == other.hash


@dataclass
class FileArtifact(Artifact):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._path: Path = Path(self.location)
        if self._path.exists() and not self._path.is_dir():
            self.hash = self.artifact_hash()

    def artifact_hash(self):
        return file_hash(self._path).hexdigest()



from dataclasses import dataclass, field
from enum import Enum
from functools import wraps
from inspect import signature
from typing import Callable, List, Dict, Optional


class ParameterType(int, Enum):

    PIPELINE_RESULTS = 1
    EXPLICIT = 2


class ResultType(str, Enum):

    VALUE = 'values'
    ARTIFACT = 'artifacts'


@dataclass
class ResultSpec:

    result_task_name: str
    result_type: ResultType
    result_var_name: str


@dataclass
class ParameterSpec:

    param_name: str
    param_type: ParameterType
    result_spec: Optional[ResultSpec] = None
    selector: Optional[Callable] = None


@dataclass
class TaskDef:

    name: str
    depends_on: Optional[List[str]]
    pure: bool
    param_specs: List[ParameterSpec] = field(default_factory=list)


class InvalidTaskDefinitionError(Exception):
    pass


def build_parameter_spec(func, selectors: Dict[str, Callable] = None):

    sig = signature(func)
    param_names = list(sig.parameters.keys())

    # three options available:
    # 1. a single parameter which will receive the full intermediate pipeline state
    # 2. any number of parameters annotated with a string of the form:
    #   '<task_name>__<values|artifacts>__<value_name|artifact_name>'
    # 3. any number of parameters accompanied by a selector dictionary, of the form:
    #   {parameter_name: callable} where the callable takes the previous state as a
    #   parameter and produces an arbitrary value
    # note the double underbars like in the django query language

    # TODO: fix bug where a single parameter cannot be underbar-referenced

    err_format = '<task_name>__<values|artifacts>__<value_name|artifact_name>'

    if len(param_names) == 0:
        spec = []
    elif len(param_names) == 1 and '__' not in param_names[0] and not selectors:
        spec = [ParameterSpec(param_names[0], ParameterType.PIPELINE_RESULTS)]
    else:
        spec = []
        for name in param_names:
            param = sig.parameters[name]
            if selectors:
                selector = selectors[name]
                spec.append(ParameterSpec(name, ParameterType.EXPLICIT, None, selector))
            elif isinstance(param.annotation, str):
                annot = param.annotation.split('__')
                if len(annot) != 3:
                    raise InvalidTaskDefinitionError(
                        f'Invalid function annotation for parameter {name}.'
                        f'Function parameters must be annotated using the following format:'
                        f'\n{err_format}')

                spec.append(ParameterSpec(name, ParameterType.EXPLICIT, ResultSpec(*annot)))
            else:
                raise InvalidTaskDefinitionError(
                    f'Annotation string or selector missing for variable {name}.'
                    f'Function parameters must be annotated using the following format:'
                    f'\n{err_format}')

    return spec


def task(_func=None, *, depends_on: Optional[List[str]] = None, pure: bool = True, selectors=None):

    def decorator_task(func: Callable):

        @wraps(func)
        def task_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        setattr(task_wrapper, 'task_def', TaskDef(
            name=func.__name__,
            depends_on=depends_on,
            pure=pure,
            param_specs=build_parameter_spec(func, selectors)
        ))

        setattr(task_wrapper, '_yenta_task', True)

        return task_wrapper

    if _func is None:
        return decorator_task
    else:
        return decorator_task(_func)



import os

from pathlib import Path


YENTA_STORE_PATH = Path('./.yenta_cache').resolve()
YENTA_ENTRY_POINT = os.environ.get('YENTA_ENTRY_POINT', Path('./main.py'))
YENTA_CONFIG_FILE = os.environ.get('YENTA_CONFIG_FILE', Path('./yenta.config'))
YENTA_LOG_FILE = os.environ.get('YENTA_LOG_FILE', None)

VERBOSE = False




from hashlib import sha1
from pathlib import Path


def file_hash(path: Path, block_size=65536):
    s = sha1()

    if path.exists():
        with open(path, 'rb') as f:
            stop = False
            while not stop:
                data = f.read(block_size)
                if len(data) > 0:
                    s.update(data)
                else:
                    stop = True

    return s



#!/usr/bin/env python

"""Tests for `yenta` package."""

import pytest
import json
import shutil

from pathlib import Path
from click.testing import CliRunner

from yenta import cli
from yenta.tasks import task
from yenta.pipeline import Pipeline, PipelineResult
from yenta.config import settings


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


@pytest.fixture
def store_path(monkeypatch):

    monkeypatch.setattr(settings, 'YENTA_STORE_PATH', Path('tests/tmp/pipeline'))
    yield settings.YENTA_STORE_PATH
    for path in settings.YENTA_STORE_PATH.iterdir():
        shutil.rmtree(path)


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.yenta)
    assert result.exit_code == 0
    for cmd in ['dump-task-graph', 'list-tasks', 'rm', 'run', 'show-config', 'task-info']:
        assert cmd in result.output


def test_list_tasks(store_path):

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', store_path, 'list-tasks'])

    output_lines = result.output.split('\n')
    assert result.exit_code == 0
    assert output_lines[0] == 'The following tasks are available:'
    assert output_lines[1] == '[ ] foo'
    assert output_lines[2] == '[ ] bar'
    assert output_lines[3] == '[ ] baz'


def test_run_tasks(store_path):

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', store_path, 'run'])

    assert result.exit_code == 0

    # bar errors, foo runs, baz never gets called
    output_lines = result.output.split('\n')
    ind = output_lines.index('[\u2718] bar')
    assert ind >= 0
    assert output_lines[ind + 1] == 'hello from foo task'
    assert output_lines[ind + 2] == '[\u2714] foo'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', store_path, 'run'])

    # bar still errors, foo recycles the old value, baz still not called
    output_lines = result.output.split('\n')
    ind = output_lines.index('[\u2718] bar')
    assert ind >= 0
    assert output_lines[ind + 1] == '[\u2014] foo'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', store_path, 'list-tasks'])

    output_lines = result.output.split('\n')
    ind = output_lines.index('The following tasks are available:')
    assert output_lines[ind + 1] == '[\u2714] foo'
    assert output_lines[ind + 2] == '[\u2718] bar'
    assert output_lines[ind + 3] == '[ ] baz'


def test_show_config(store_path):

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point, '--pipeline', store_path, 'show-config'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'Yenta is using the following configuration:'
    assert output_lines[1] == f'The entrypoint for Yenta is {entry_point}'
    assert output_lines[2] == f'Pipelines will be cached in {store_path}'
    assert output_lines[3] == 'No log output configured'
    print(result.output)


def test_task_info(store_path):

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', store_path,
                                       'task-info', 'nonexistent-task'])

    assert result.exit_code == 0
    assert result.output == 'Unknown task nonexistent-task specified.\n'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', store_path,
                                       'run'])

    assert result.exit_code == 0
    assert Path(store_path / 'default').exists()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', store_path,
                                       'task-info', 'foo'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'Information for task foo:'
    assert output_lines[1] == 'Dependencies: None'
    assert output_lines[2] == 'Previous status: \u2714'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', store_path,
                                       'task-info', 'bar'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'Information for task bar:'
    assert output_lines[1] == 'Dependencies: None'
    assert output_lines[2] == 'Previous status: \u2718 oh noes'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', store_path,
                                       'task-info', 'baz'])

    output_lines = result.output.split('\n')
    assert output_lines[0] == 'Information for task baz:'
    assert output_lines[1] == 'Dependencies: foo, bar'
    assert output_lines[2] == 'Previous status: Did not run'


def test_rm_task(store_path):

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', store_path,
                                       'run'])

    assert result.exit_code == 0
    assert Path(store_path / 'default').exists()

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', store_path,
                                       'rm', 'foo'])

    assert result.exit_code == 0
    assert Path(store_path / 'default').exists()

    pipeline = Pipeline.load_pipeline(store_path / 'default')

    with pytest.raises(KeyError) as ex:
        _ = pipeline.values('foo', 'whatever')

    assert 'foo' in str(ex.value)

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', store_path,
                                       'rm', 'nonexistent-task'])

    assert result.exit_code == 0
    assert result.output == 'Unknown task nonexistent-task specified.\n'


def test_dump_task_graph():

    runner = CliRunner()
    entry_point = 'tests/sample_pipelines/sample_pipeline_1.py'
    pipeline_store = 'tests/sample_pipelines/sample_pipeline_1.json'
    if Path(pipeline_store).exists():
        Path(pipeline_store).unlink()

    task_graph = 'tests/sample_pipelines/sample_task_graph.png'

    result = runner.invoke(cli.yenta, ['--entry-point', entry_point,
                                       '--pipeline', pipeline_store,
                                       'dump-task-graph', task_graph])

    assert result.exit_code == 0
    assert Path(task_graph).exists()

    Path(task_graph).unlink()



#!/usr/bin/env python3
"""Console script for yenta."""
import sys
import click
import configparser
import importlib.util
import more_itertools
import shutil
import os

from rich.tree import Tree
from rich.text import Text
from rich import print

from typing import Iterable
from networkx.drawing.nx_pydot import to_pydot
from colorama import init, Fore, Style
from pathlib import Path
from yenta.config import settings
from yenta.pipeline.Pipeline import Pipeline, TaskStatus

import logging

logger = logging.getLogger(__name__)


CHECK_MARK = u'\u2714'
X_MARK = u'\u2718'


def load_tasks(entry_file):
    spec = importlib.util.spec_from_file_location('main', entry_file)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    tasks = [func for _, func in module.__dict__.items()
             if callable(func) and hasattr(func, '_yenta_task')]

    return tasks


@click.group()
@click.option('--config-file', default=settings.YENTA_CONFIG_FILE, type=Path,
              help='The config file from which to read settings.')
@click.option('--pipeline-store', type=Path, help='The directory to which the pipeline will be cached.')
@click.option('--entry-point', type=Path, help='The file containing the task definitions.')
@click.option('--log-file', type=Path, help='The file to which the logs should be written.')
def yenta(config_file, pipeline_store, entry_point, log_file):

    init()

    # append the local path we're running from so that we can allow
    # the project to import normally when running via CLI
    sys.path.append(os.getcwd())

    cf = configparser.ConfigParser()
    cf.read(config_file or settings.YENTA_CONFIG_FILE)
    if 'yenta' not in cf:
        cf['yenta'] = {}

    settings.YENTA_ENTRY_POINT = entry_point or \
                                 cf['yenta'].get('entry_point', None) or \
                                 settings.YENTA_ENTRY_POINT

    pipeline_file = cf['yenta'].get('pipeline_store', None)
    pipeline_path = Path(pipeline_file).resolve() if pipeline_file else None
    settings.YENTA_STORE_PATH = pipeline_store or \
                                pipeline_path or \
                                settings.YENTA_STORE_PATH
    conf_log_file = cf['yenta'].get('log_file', None)
    conf_log_path = Path(conf_log_file).resolve() if log_file else None
    settings.YENTA_LOG_FILE = log_file or \
                              conf_log_path or \
                              settings.YENTA_LOG_FILE


@yenta.command(help='List all available tasks.')
@click.option('--pipeline-name', default='default', help='The name of the pipeline to display.')
def list_tasks(pipeline_name='default'):

    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    pipeline_data = Pipeline.load_pipeline(settings.YENTA_STORE_PATH / pipeline_name)

    print('[bold white]The following tasks are available:[/bold white]')
    for task in tasks:
        task_name = task.task_def.name
        task_result = pipeline_data.task_results.get(task_name, None)
        marker = ' '
        if task_result and task_result.status == TaskStatus.SUCCESS:
            marker = f'[bold green]{CHECK_MARK}[/bold green]'
        elif task_result and task_result.status == TaskStatus.FAILURE:
            marker = f'[bold red]{X_MARK}[/bold red]'

        print(f'[{marker}] [bold white]{task_name}[/bold white]')


@yenta.command(help='Show the current configuration.')
def show_config():

    print('[bold white]Yenta is using the following configuration:[/bold white]')
    print(f'[bold white]The entrypoint for Yenta is [green]{settings.YENTA_ENTRY_POINT}[/green][/bold white]')
    print(f'[bold white]Pipelines will be cached in [green]{settings.YENTA_STORE_PATH}[/green][/bold white]')
    if settings.YENTA_LOG_FILE:
        print('Log output will be written to ' + Fore.GREEN + str(settings.YENTA_LOG_FILE) + Fore.WHITE)
    else:
        print('No log output configured')


@yenta.command(help='Show information about a specific task.')
@click.argument('task-name')
@click.option('--pipeline-name', default='default', help='The name of the pipeline to display.')
def task_info(task_name, pipeline_name='default'):

    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    pipeline_data = Pipeline.load_pipeline(settings.YENTA_STORE_PATH / pipeline_name)
    try:
        task = more_itertools.one(filter(lambda t: t.task_def.name == task_name, tasks))
        print(f'[bold white]Information for task [green]{task_name}[/green]:[/bold white]')
        deps = ', '.join(task.task_def.depends_on) if task.task_def.depends_on else 'None'
        print(f'[bold white]Dependencies: [bright_blue]{deps}[/bright_blue][/bold white]')
        task_result = pipeline_data.task_results.get(task_name, None)
        marker = 'Did not run'
        if task_result and task_result.status == TaskStatus.SUCCESS:
            marker = Fore.GREEN + u'\u2714' + Fore.WHITE
        elif task_result and task_result.status == TaskStatus.FAILURE:
            marker = Fore.RED + u'\u2718' + ' ' + task_result.error + Fore.WHITE
        print(f'Previous status: {marker}')

        tree = Tree('Previous result: ')
        if task_result and task_result.status == TaskStatus.SUCCESS:
            # print('Previous result: ', task_result)
            values_node = tree.add('values')
            for key in sorted(task_result.values.keys()):
                val = task_result.values.get(key)
                if isinstance(val, Iterable) and not isinstance(val, str):
                    key_node = values_node.add(Text(f'{key}: '))
                    for v in val:
                        key_node.add(Text(str(v)))
                else:
                    values_node.add(Text(f'{key}: {val}'))
            artifacts_node = tree.add('artifacts')
            for key in sorted(task_result.artifacts.keys()):
                val = task_result.artifacts.get(key)
                if isinstance(val, Iterable) and not isinstance(val, str):
                    key_node = artifacts_node.add(Text(f'{key}: '))
                    for v in val:
                        key_node.add(Text(v))
                else:
                    artifacts_node.add(Text(f'{key}: {val}'))
            print(tree)
        else:
            print('Previous result: ' + Fore.GREEN + 'None' + Fore.WHITE)

    except ValueError:
        print(Fore.WHITE + Style.BRIGHT + 'Unknown task ' + Fore.RED + task_name + Fore.WHITE + ' specified.')


@yenta.command(help='Remove a task from the pipeline cache.')
@click.argument('task-name')
@click.option('--pipeline-name', default='default', help='The name of the pipeline to display.')
def rm(task_name, pipeline_name='default'):

    task_path = settings.YENTA_STORE_PATH / pipeline_name / task_name

    if task_path.exists():
        shutil.rmtree(task_path)
    else:
        print(Fore.WHITE + Style.BRIGHT + 'Unknown task ' + Fore.RED + task_name + Fore.WHITE + ' specified.')


@yenta.command(help='Mark a task as ignorable; it will be skipped by the pipeline.')
@click.argument('task-name')
@click.option('--pipeline-name', default='default', help='The name of the pipeline to display.')
def ignore(task_name, pipeline_name='default'):

    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    target_task = [task for task in tasks if task.task_def.name == task_name]
    if len(target_task) == 0:
        print(f'[bold white]Unknown task {task_name} specified.[/bold white]')
    elif len(target_task) > 1:
        print(f'[bold red]Multiple tasks have the same name: {task_name}. Task names must be unique '
              f'within a pipeline.[/bold red]')
    else:
        ignore_path = settings.YENTA_STORE_PATH / pipeline_name / task_name / '.ignore'
        ignore_path.touch(exist_ok=True)
        print(f'[bold white]Setting task {task_name} to be ignored.[/bold white]')


@yenta.command(help='Dump the task graph to a DOT file.')
@click.argument('filename', type=click.Path())
def dump_task_graph(filename: Path):

    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    pipeline = Pipeline(*tasks)
    pydot_graph = to_pydot(pipeline.task_graph)
    pydot_graph.write(filename)


@yenta.command(help='Run the pipeline.')
@click.option('--up-to', help='Optionally run the pipeline up to and including a given task.')
@click.option('--force-rerun', '-f', multiple=True, default=[], help='Force specified tasks to rerun.')
@click.option('--pipeline-name', default='default', help='The name of the pipeline to run.')
def run(up_to=None, force_rerun=None, pipeline_name='default'):

    logger.info('Running the pipeline')
    tasks = load_tasks(settings.YENTA_ENTRY_POINT)
    pipeline = Pipeline(*tasks, name=pipeline_name)
    result = pipeline.run_pipeline(up_to, force_rerun)


if __name__ == "__main__":
    sys.exit(yenta())  # pragma: no cover
